from typing import Optional

import numpy as np
import numpy.typing as npt


def softmax(x: npt.NDArray[np.float32], axis=1) -> npt.NDArray[np.float32]:
    exp_x = np.exp(x - np.max(x, axis=axis, keepdims=True))
    return exp_x / np.sum(exp_x, axis=axis, keepdims=True)


def sigmoid(x: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
    return (1 / (1 + np.exp(-x))).astype(np.float32)


def gaussian_weights(
    size: int, sigma: Optional[float] = None
) -> npt.NDArray[np.float32]:
    """
    Generates a 1D Gaussian weight array.

    Args:
        size (int): Size of the array.
        sigma (float): Standard deviation of the Gaussian distribution.

    Returns:
        numpy.ndarray: 1D array of Gaussian weights.
    """
    if sigma is None:
        sigma = size / 4
    vals = np.arange(size) - size // 2
    x, y = np.meshgrid(vals, vals)
    weights = np.exp(-0.5 * (x**2 + y**2) / sigma**2)
    return (weights / np.max(weights)).astype(np.float32)


def _get_pad(size: int, tile_size: int, stride: int) -> int:
    if size < tile_size:
        return tile_size - size
    remainder = (size - tile_size) % stride
    if remainder == 0:
        return 0
    return stride - remainder


def get_padding(size: int, tile_size: int, stride: int) -> tuple[int, int]:
    pad = _get_pad(size, tile_size, stride)
    if pad == 0:
        return 0, 0
    pad_before = pad // 2
    pad_after = pad - pad_before
    return pad_before, pad_after



from typing import TYPE_CHECKING, Optional

import numpy as np
import numpy.typing as npt

from .tile_segmenter import TileSegmenterBase
from .segmentation_result import SegmentationResult
from .utils import gaussian_weights, get_padding
from .types import ActivationType, BlendModeType
from .defaults import (
    DEFAULT_ACTIVATION,
    DEFAULT_NUM_CLASSES,
    DEFAULT_TILE_SIZE,
    DEFAULT_OVERLAP,
    DEFAULT_BLEND_MODE,
    DEFAULT_PAD_VALUE,
    DEFAULT_BATCH_SIZE,
)

if TYPE_CHECKING:
    import torch
    import onnxruntime as ort  # type: ignore[import-untyped]


class Segmenter:
    def __init__(
        self,
        tile_segmenter: TileSegmenterBase,
        *,
        num_classes: int = DEFAULT_NUM_CLASSES,
        tile_size: int = DEFAULT_TILE_SIZE,
        overlap: float = DEFAULT_OVERLAP,
        blend_mode: BlendModeType = DEFAULT_BLEND_MODE,
        pad_value: int = DEFAULT_PAD_VALUE,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> None:
        assert num_classes > 0, "Number of classes must be greater than 0"
        self.num_classes = num_classes
        self.tile_segmenter = tile_segmenter
        if blend_mode == "gaussian":
            # Expand dims to [H, W, 1] to broadcast across classes
            self.blend_weights = gaussian_weights(tile_size)[..., None]
        else:
            self.blend_weights = np.ones((tile_size, tile_size, 1), dtype=np.float32)
        self.tile_size = tile_size
        self.stride = round(tile_size * (1 - overlap))
        self.pad_value = pad_value
        self.batch_size = batch_size

    def __call__(self, image: npt.NDArray[np.uint8]) -> SegmentationResult:
        return SegmentationResult(self._segment(image))

    def _segment(self, image: npt.NDArray[np.uint8]) -> npt.NDArray[np.float32]:
        assert image.ndim == 3, "Input image must have 3 dimensions (H x W x C)"
        H, W, _C = image.shape
        pad_y = get_padding(H, self.tile_size, self.stride)
        pad_x = get_padding(W, self.tile_size, self.stride)
        image_padded = np.pad(
            image,
            (pad_y, pad_x, (0, 0)),
            mode="constant",
            constant_values=self.pad_value,
        )
        probs_padded = self._segment_padded(image_padded)
        probs = probs_padded[pad_y[0] : -pad_y[1], pad_x[0] : -pad_x[1]]
        return probs

    def _segment_padded(self, image: npt.NDArray[np.uint8]) -> npt.NDArray[np.float32]:
        H, W, _C = image.shape

        # Initialize outputs with class dimension
        output_probs = np.zeros((H, W, self.num_classes), dtype=np.float32)
        output_weights = np.zeros((H, W, self.num_classes), dtype=np.float32)

        # Generate tile coordinates
        tile_coords = [
            (y, x)
            for y in range(0, H - self.tile_size + 1, self.stride)
            for x in range(0, W - self.tile_size + 1, self.stride)
        ]

        for tile_coords_batch in [
            tile_coords[i : i + self.batch_size]
            for i in range(0, len(tile_coords), self.batch_size)
        ]:
            # Extract tiles from image (N x H x W x C)
            tiles = np.stack(
                [
                    image[y : y + self.tile_size, x : x + self.tile_size]
                    for y, x in tile_coords_batch
                ]
            )
            # Get predictions (N x H x W x num_classes)
            outputs = self.tile_segmenter(tiles)

            for (y, x), output in zip(tile_coords_batch, outputs):
                output_probs[y : y + self.tile_size, x : x + self.tile_size] += (
                    output
                    * self.blend_weights  # blend_weights broadcasts to [H,W,num_classes]
                )
                output_weights[y : y + self.tile_size, x : x + self.tile_size] += (
                    self.blend_weights
                )

        # Blend probabilities
        probs = np.divide(
            output_probs,
            output_weights,
            out=np.zeros_like(output_probs, dtype=np.float32),
            where=output_weights != 0,
        ).astype(np.float32)
        return probs

    @staticmethod
    def from_torch(
        model: "torch.nn.Module",
        *,
        activation: ActivationType = DEFAULT_ACTIVATION,
        device: Optional["torch.device"] = None,
        num_classes: int = DEFAULT_NUM_CLASSES,
        tile_size: int = DEFAULT_TILE_SIZE,
        overlap: float = DEFAULT_OVERLAP,
        blend_mode: BlendModeType = DEFAULT_BLEND_MODE,
        pad_value: int = DEFAULT_PAD_VALUE,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> "Segmenter":
        from .torch import TorchTileSegmenter

        tile_segmenter = TorchTileSegmenter(model, activation=activation, device=device)
        return Segmenter(
            tile_segmenter,
            num_classes=num_classes,
            tile_size=tile_size,
            overlap=overlap,
            blend_mode=blend_mode,
            pad_value=pad_value,
            batch_size=batch_size,
        )

    @staticmethod
    def from_onnx(
        session: "ort.InferenceSession",
        *,
        activation: ActivationType = DEFAULT_ACTIVATION,
        num_classes: int = DEFAULT_NUM_CLASSES,
        tile_size: int = DEFAULT_TILE_SIZE,
        overlap: float = DEFAULT_OVERLAP,
        blend_mode: BlendModeType = DEFAULT_BLEND_MODE,
        pad_value: int = DEFAULT_PAD_VALUE,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> "Segmenter":
        from .onnx import OnnxTileSegmenter

        tile_segmenter = OnnxTileSegmenter(session, activation=activation)
        return Segmenter(
            tile_segmenter,
            num_classes=num_classes,
            tile_size=tile_size,
            overlap=overlap,
            blend_mode=blend_mode,
            pad_value=pad_value,
            batch_size=batch_size,
        )



from typing import cast

import numpy as np
import numpy.typing as npt
import onnxruntime as ort  # type: ignore[import-untyped]

from .tile_segmenter import TileSegmenterBase
from .utils import sigmoid, softmax
from .types import ActivationType, TilesTransformType
from .defaults import DEFAULT_ACTIVATION, DEFAULT_TILES_TRANSFORM


class OnnxTileSegmenter(TileSegmenterBase):
    def __init__(
        self,
        session: ort.InferenceSession,
        *,
        transform: TilesTransformType = DEFAULT_TILES_TRANSFORM,
        activation: ActivationType = DEFAULT_ACTIVATION,
    ) -> None:
        """
        Wraps a ONNX model for segmentation.
        We assume that given an input of shape (N, C, H, W), the model returns an output of shape (N, K, H, W).
        If the model outputs logits, then you must set `activation="sigmoid"` for binary or multiclass segmentation or
        `activation="softmax"` for multiclass.
        If `activation=None`, we assume the model has already normalized the outputs.
        """
        super().__init__(transform)
        self.session = session
        self.activation = activation
        assert len(self.session.get_inputs()) == 1, "Model must have exactly 1 input"
        self.input_name = self.session.get_inputs()[0].name
        assert len(self.session.get_outputs()) >= 1, "Model must have at least 1 output"

    def segment(self, tiles: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        ort_inputs = {self.input_name: tiles}
        output = cast(
            npt.NDArray[np.float32],
            self.session.run(output_names=None, input_feed=ort_inputs)[0],
        )
        if self.activation == "sigmoid":
            output = sigmoid(output)
        elif self.activation == "softmax":
            output = softmax(output, axis=1)
        return output



from typing import Callable, Literal, Optional

import numpy as np
import numpy.typing as npt

ActivationType = Optional[Literal["sigmoid", "softmax"]]
BlendModeType = Literal["gaussian", "flat"]
TilesTransformType = Callable[[npt.NDArray[np.uint8]], npt.NDArray[np.float32]]



from typing import Sequence

import numpy as np
import numpy.typing as npt

from .defaults import DEFAULT_THRESHOLD


class SegmentationResult:
    def __init__(
        self,
        probabilities: npt.NDArray[np.float32],
    ) -> None:
        # Shape: [H, W, C]
        self.probabilities = probabilities

    def to_binary(self) -> "BinarySegmentationResult":
        assert self.probabilities.shape[2] == 1, (
            "Model must have exactly 1 output channel"
        )
        return BinarySegmentationResult(self.probabilities[:, :, 0])

    def to_multiclass(self) -> "MulticlassSegmentationResult":
        assert self.probabilities.shape[2] > 1, (
            "Model must have more than 1 output channel"
        )
        return MulticlassSegmentationResult(self.probabilities)

    def to_multilabel(self) -> "MultilabelSegmentationResult":
        assert self.probabilities.shape[2] > 1, (
            "Model must have more than 1 output channel"
        )
        return MultilabelSegmentationResult(self.probabilities)


class BinarySegmentationResult:
    def __init__(
        self,
        probabilities: npt.NDArray[np.float32],
    ) -> None:
        self.probabilities = probabilities

    def get_mask(self, threshold: float = DEFAULT_THRESHOLD) -> npt.NDArray[np.bool_]:
        """
        Returns a H x W boolean mask `out` where `out[h,w]` indicates whether the pixel at (h, w)
        is in the positive class.
        """
        return self.probabilities > threshold

    def get_mask_proba(self, copy=True) -> npt.NDArray[np.float32]:
        """
        Returns a H x W array `out` where `out[h,w]` indicates the "probability"
        that the pixel at (h, w) is in the positive class.
        """
        if copy:
            return self.probabilities.copy()
        else:
            return self.probabilities


class MulticlassSegmentationResult:
    def __init__(
        self,
        probabilities: npt.NDArray[np.float32],  # Shape: [H, W, C]
    ) -> None:
        self.probabilities = probabilities

    def get_mask(self) -> npt.NDArray[np.uint]:
        """
        Returns a H x W array `out` where `out[h,w]` indicates the predicted class
        for the pixel at (h, w).
        """
        return np.argmax(self.probabilities, axis=2).astype(np.uint)

    def get_mask_proba(self, copy=True) -> npt.NDArray[np.float32]:
        """
        Returns a H x W x C array `out` where `out[h,w,c]` indicates the "probability"
        that the pixel at (h, w) is in class c.
        """
        if copy:
            return self.probabilities.copy()
        else:
            return self.probabilities


class MultilabelSegmentationResult:
    def __init__(
        self,
        probabilities: npt.NDArray[np.float32],  # Shape: [H, W, C]
    ) -> None:
        self.probabilities = probabilities

    def get_mask(
        self, threshold: float | Sequence[float] = DEFAULT_THRESHOLD
    ) -> npt.NDArray[np.bool_]:
        """
        Returns a H x W x C array `out` where `out[h,w,c]` whether the pixel at (h, w)
        belongs to label c.
        """
        if isinstance(threshold, float):
            return self.probabilities > threshold
        else:
            assert len(threshold) == self.probabilities.shape[2], (
                "Threshold must have the same length as number of classes"
            )
            return self.probabilities > np.array(threshold)[None, None, :]

    def get_mask_proba(self, copy=True) -> npt.NDArray[np.float32]:
        """
        Returns a H x W x C array `out` where `out[h,w,c]` indicates the "probability"
        that the pixel at (h, w) has label c
        """
        if copy:
            return self.probabilities.copy()
        else:
            return self.probabilities



from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt

from .types import TilesTransformType
from .defaults import DEFAULT_TILES_TRANSFORM


class TileSegmenterBase(ABC):
    def __init__(self, transform: TilesTransformType = DEFAULT_TILES_TRANSFORM):
        self.transform = transform

    @abstractmethod
    def segment(self, tiles: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        pass

    def __call__(self, tiles: npt.NDArray[np.uint8]) -> npt.NDArray[np.float32]:
        """
        Assumes tiles is N x C x H x W
        Returns an array of shape N x K x H x W
        """
        assert tiles.ndim == 4, (
            "Input image tiles must have 4 dimensions (N x C x H x W)"
        )
        N, H, W, _C = tiles.shape
        # assert C == 3, "Input image tiles must have 3 channels (RGB)"
        # Move channel dimension (N x H x W x C => N x C x H x W)
        tiles = np.transpose(tiles, (0, 3, 1, 2))
        # Output shape is N x K x H x W
        output = self.segment(self.transform(tiles))
        assert output.ndim == 4, "Output mask must have 4 dimensions (N x K x H x W)"
        assert output.shape[0] == N, (
            "Output mask must have the same number of samples as the input image"
        )
        assert output.shape[2:] == (H, W), (
            "Output mask must have the same shape as the input image"
        )
        # Move class dimension (N x K x H x W => N x H x W x K)
        output = np.transpose(output, (0, 2, 3, 1))
        return output



import numpy as np
import numpy.typing as npt

from .types import ActivationType, BlendModeType

DEFAULT_NUM_CLASSES: int = 1
DEFAULT_TILE_SIZE: int = 512
DEFAULT_OVERLAP: float = 0.125
DEFAULT_BLEND_MODE: BlendModeType = "gaussian"
DEFAULT_PAD_VALUE: int = 255
DEFAULT_BATCH_SIZE: int = 1
DEFAULT_ACTIVATION: ActivationType = None
DEFAULT_THRESHOLD: float = 0.5


def DEFAULT_TILES_TRANSFORM(tiles: npt.NDArray[np.uint8]):
    return (tiles / 255).astype(np.float32)



from typing import cast, Optional

import numpy as np
import numpy.typing as npt
import torch

from .tile_segmenter import TileSegmenterBase
from .types import ActivationType, TilesTransformType
from .defaults import DEFAULT_ACTIVATION, DEFAULT_TILES_TRANSFORM


class TorchTileSegmenter(TileSegmenterBase):
    def __init__(
        self,
        model: torch.nn.Module,
        *,
        transform: TilesTransformType = DEFAULT_TILES_TRANSFORM,
        activation: ActivationType = DEFAULT_ACTIVATION,
        device: Optional[torch.device],
    ) -> None:
        """
        Wraps a PyTorch model for segmentation.
        We assume that given an input of shape (N, C, H, W), the model returns an output of shape (N, K, H, W).
        If the model outputs logits, then you must set `activation="sigmoid"` for binary or multiclass segmentation or
        `activation="softmax"` for multiclass.
        If `activation=None`, we assume the model has already normalized the outputs.
        """
        super().__init__(transform)
        self.model = model.eval()
        self.activation = activation
        self.device = device if device else torch.get_default_device()
        self.model.to(self.device)

    def segment(self, tiles: npt.NDArray[np.float32]) -> npt.NDArray[np.float32]:
        x = torch.from_numpy(tiles).to(self.device)
        with torch.inference_mode():
            output = self.model(x)
            if self.activation == "sigmoid":
                output = output.sigmoid()
            elif self.activation == "softmax":
                output = output.softmax(dim=1)
            return cast(npt.NDArray[np.float32], output.cpu().numpy())
